from __future__ import absolute_import
from __future__ import print_function

import six
import pickle
import glob
import os
import sys
import subprocess
import numpy as np
import yaml
from collections import OrderedDict
from contextlib import contextmanager
import inspect
import logging
import collections
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def load_module(path, module_name=None):
    """Load python module from file

    Args:
       path: python file path
       module_name: import as `module_name` name. If none, use `path[:3]`
    """
    assert path.endswith(".py")
    if module_name is None:
        module_name = os.path.basename(path)[:-3]  # omit .py

    logger.debug("loading module: {0} as {1}".format(path, module_name))
    if sys.version_info[0] == 2:
        import imp
        # add the directory to system's path - allows loading submodules
        sys.path.append(os.path.join(os.path.dirname(module_name)))
        module = imp.load_source(module_name, path)
    elif sys.version_info[0] == 3:
        # TODO: implement dynamic loading of preprocessor module for python3
        """
        import importlib.machinery
        loader = importlib.machinery.SourceFileLoader
        handle = loader.load_module
        """
        if sys.version_info[1] == 4:
            # way 1 (=python3.4)
            from importlib.machinery import SourceFileLoader
            module = SourceFileLoader(module_name, path).load_module()
        if sys.version_info[1] >= 5:
            # way 2 (>=python3.5)
            import importlib.util

            # alternative
            import types
            loader = importlib.machinery.SourceFileLoader(module_name, path)
            module = types.ModuleType(loader.name)
            loader.exec_module(module)

            # module_spec_ = importlib.util.spec_from_file_location(module_name, path)
            # module = importlib.util.module_from_spec(module_spec_)
            # module_spec_.loader.exec_module(module)
        else:
            raise RuntimeError(
                'dynamic loading of preprocessor module is not implemented for python3!')
    return module


def pip_install_requirements(requirements_fname):
    if os.path.exists(requirements_fname):  # install dependencies
        logger.info('Running pip install -r {}...'.format(requirements_fname))
        subprocess.call(['pip', 'install', '-r', requirements_fname])
    else:
        logger.info('requirements.txt not found under {}'.format(requirements_fname))


def compare_numpy_dict(a, b, exact=True):
    """
    Compare two recursive numpy dictionaries or lists
    """
    if type(a) != type(b) and type(a) != np.ndarray and type(b) != np.ndarray:
        return False

    # Compare two dictionaries
    if type(a) == dict and type(b) == dict:
        if not a.keys() == b.keys():
            return False
        for key in a.keys():
            res = compare_numpy_dict(a[key], b[key], exact)
            if not res:
                print("false for key = ", key)
                return False
        return True

    # compare two lists
    if type(a) == list and type(b) == list:
        assert len(a) == len(b)
        return all([compare_numpy_dict(a[i], b[i], exact=exact)
                    for i in range(len(a))])

    # if type(a) == np.ndarray and type(b) == np.ndarray:
    if type(a) == np.ndarray or type(b) == np.ndarray:
        if exact:
            return (a == b).all()
        else:
            return np.testing.assert_almost_equal(a, b)

    if a is None and b is None:
        return True

    raise NotImplementedError


def parse_json_file_str(extractor_args):
    """Parse a string either as a json string or
    as a file path to a .json file
    """
    extractor_args = extractor_args.strip("'").strip('"')
    if extractor_args.startswith("{") or extractor_args.endswith("}"):
        logger.debug("Parsing the extractor_args as a json string")
        return yaml.load(extractor_args)
    else:
        if not os.path.exists(extractor_args):
            raise ValueError("File path: {0} doesn't exist".format(extractor_args))
        logger.debug("Parsing the extractor_args as a json file path")
        with open(extractor_args, "r") as f:
            return yaml.load(f.read())


# https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts


def yaml_ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict):
    class OrderedLoader(Loader):
        pass

    def dict_constructor(loader, node):
        return OrderedDict(loader.construct_pairs(node))

    _mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG
    OrderedLoader.add_constructor(_mapping_tag, dict_constructor)
    return yaml.load(stream, OrderedLoader)


def yaml_ordered_dump(data, stream=None, Dumper=yaml.Dumper, **kwds):
    class OrderedDumper(Dumper):
        pass

    def dict_representer(dumper, data):
        return dumper.represent_dict(six.iteritems(data))
    OrderedDumper.add_representer(OrderedDict, dict_representer)
    return yaml.dump(data, stream, OrderedDumper, **kwds)


@contextmanager
def cd(newdir):
    """Temporarily change the directory
    """
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def getargs(x):
    """Get function arguments
    """
    if sys.version_info[0] == 2:
        if inspect.isfunction(x):
            return set(inspect.getargspec(x).args)
        else:
            # skip the self parameter
            return set(inspect.getargspec(x.__init__).args[1:])
    else:
        return set(inspect.signature(x).parameters.keys())


def read_yaml(path):
    with open(path) as f:
        return yaml.load(f)


def cmd_exists(cmd):
    """Check if a certain command exists
    """
    return subprocess.call("type " + cmd, shell=True,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE) == 0


def lfs_installed(raise_exception=False):
    """Check if git lfs is installed localls
    """
    ce = cmd_exists("git-lfs")
    if not ce:
        if raise_exception:
            raise OSError("git-lfs not installed")
        else:
            logger.warn("git-lfs not installed")
    return ce


def get_file_path(file_dir, basename, extensions=[".yml", ".yaml"],
                  raise_err=True):
    """Get the file path allowing for multiple file extensions
    """
    for ext in extensions:
        path = os.path.join(file_dir, basename + ext)
        if os.path.exists(path):
            return path
    if raise_err:
        raise ValueError("File path doesn't exists: {0}/{1}{2}".
                         format(file_dir, basename, set(extensions)))
    else:
        return None


def du(path):
    """disk usage in human readable format (e.g. '2,1GB')"""
    try:
        return subprocess.check_output(['du', '-sh', path]).split()[0].decode('utf-8')
    except Exception:
        return "NA"


class Slice_conv:

    def __getitem__(self, key):
        return key


def unique_list(seq):
    """Make a list unique and preserve the elements order

    Modified version of Dave Kirby solution
    """
    seen = set()
    return [x for x in seq if x not in seen and not seen.add(x)]


def read_txt(file_path, comment_str="#"):
    """Txt file reader that ignores comments and
    empty lines
    """
    out = []
    with open(file_path) as f:
        for line in f:
            line = line.partition(comment_str)[0]
            line = line.strip()
            if len(line) > 0:
                out.append(line)
    return out


def read_pickle(f):
    with open(f, "rb") as f:
        return pickle.load(f)


def merge_dicts(x, y):
    """https://stackoverflow.com/questions/38987/how-to-merge-two-dictionaries-in-a-single-expression
    """
    z = x.copy()   # start with x's keys and values
    z.update(y)    # modifies z with y's keys and values & returns None
    return z


def list_files_recursively(root_dir, basename, suffix='y?ml'):
    """search for filenames matching the pattern: {root_dir}/**/{basename}.{suffix}
    """
    if sys.version_info >= (3, 5):
        return [filename[len(root_dir):] for filename in
                glob.iglob(root_dir + '**/{0}.{1}'.format(basename, suffix), recursive=True)]
    else:
        import fnmatch
        return [os.path.join(root, filename)[len(root_dir):]
                for root, dirnames, filenames in os.walk(root_dir)
                for filename in fnmatch.filter(filenames, '{0}.{1}'.format(basename, suffix))]


def map_nested(dd, fn):
    """Map a function to a nested data structure (containing lists or dictionaries

    Args:
      dd: nested data structure
      fn: function to apply to each leaf
    """
    if isinstance(dd, collections.Mapping):
        return {key: map_nested(dd[key], fn) for key in dd}
    elif isinstance(dd, collections.Sequence):
        return [map_nested(x, fn) for x in dd]
    else:
        return fn(dd)


def take_first_nested(dd):
    """Get a single element from the nested list/dictionary

    Args:
      dd: nested data structure

    Example: take_first_nested({"a": [1,2,3], "b": 4}) == 1
    """
    if isinstance(dd, collections.Mapping):
        return take_first_nested(six.next(six.itervalues(dd)))
    elif isinstance(dd, collections.Sequence):
        return take_first_nested(dd[0])
    else:
        return dd



from __future__ import absolute_import
from __future__ import print_function

import os
import abc

import kipoi  # for .config module
from kipoi.components import DataLoaderDescription, example_kwargs, print_dl_kwargs
from .utils import load_module, cd, getargs
from .external.torch.data import DataLoader
from kipoi.data_utils import numpy_collate, numpy_collate_concat, get_dataset_item, get_dataset_lens, iter_cycle
from tqdm import tqdm
import types

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

#
PREPROC_IFILE_TYPES = ['DNA_regions']
PREPROC_IFILE_FORMATS = ['bed3']


class BaseDataLoader(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def batch_iter(self, **kwargs):
        raise NotImplementedError

    def batch_train_iter(self, cycle=True, **kwargs):
        """Returns samples directly useful for training the model:
        (x["inputs"],x["targets"])

        Args:
          cycle: when True, the returned iterator will run indefinitely by using
            a custom `itertools.cycle` implementation. Use True with `fit_generator` in Keras.
          **kwargs: Arguments passed to self.batch_iter(**kwargs)
        """
        if cycle:
            return ((x["inputs"], x["targets"]) for x in iter_cycle(self.batch_iter(**kwargs)))
        else:
            return ((x["inputs"], x["targets"]) for x in self.batch_iter(**kwargs))

    def batch_predict_iter(self, **kwargs):
        """Returns samples directly useful for prediction x["inputs"]

        Args:
          **kwargs: Arguments passed to self.batch_iter(**kwargs)
        """
        return (x["inputs"] for x in self.batch_iter(**kwargs))

    def load_all(self, **kwargs):
        """Loads and returns the whole dataset

        Arguments:
            **kwargs: passed to batch_iter()
        """
        return numpy_collate_concat([x for x in tqdm(self.batch_iter(**kwargs))])

# --------------------------------------------
# Different implementations

# Other options:
# - generator - sample, batch-based
#   - yield.
# - iterator, iterable - sample, batch-based
#   - __iter__()
#     - __next__()
# - full dataset
#   - everything numpy arrays with the same first axis length


class PreloadedDataset(BaseDataLoader):
    """Generated by supplying a function returning the full dataset.

    The full dataset is a nested (list/dict) python structure of numpy arrays
    with the same first axis dimension.
    """
    data_fn = None

    @classmethod
    def from_fn(cls, data_fn):
        """setup the class variable
        """
        cls.data_fn = staticmethod(data_fn)
        return cls

    @classmethod
    def from_data(cls, data):
        return cls.from_data_fn(lambda: data)()

    @classmethod
    def _get_data_fn(cls):
        assert cls.data_fn is not None
        return cls.data_fn

    def __init__(self, *args, **kwargs):
        self.data = self._get_data_fn()(*args, **kwargs)
        lens = get_dataset_lens(self.data, require_numpy=True)
        # check that all dimensions are the same
        assert len(set(lens)) == 1
        self.n = lens[0]

    def __len__(self):
        return self.n

    def __getitem__(self, index):
        return get_dataset_item(self.data, index)

    def batch_iter(self, batch_size=32, shuffle=False, drop_last=False, **kwargs):
        """Return a batch-iterator

        Arguments:
            dataset (Dataset): dataset from which to load the data.
            batch_size (int, optional): how many samples per batch to load
                (default: 1).
            shuffle (bool, optional): set to ``True`` to have the data reshuffled
                at every epoch (default: False).
            num_workers (int, optional): how many subprocesses to use for data
                loading. 0 means that the data will be loaded in the main process
                (default: 0)
            drop_last (bool, optional): set to ``True`` to drop the last incomplete batch,
                if the dataset size is not divisible by the batch size. If False and
                the size of dataset is not divisible by the batch size, then the last batch
                will be smaller. (default: False)

        Returns:
            iterator
        """
        dl = DataLoader(self, batch_size=batch_size,
                        collate_fn=numpy_collate,
                        shuffle=shuffle,
                        num_workers=0,
                        drop_last=drop_last)
        return iter(dl)

    def load_all(self, **kwargs):
        """Load the whole dataset into memory

        Arguments:
            **kwargs: ignored
        """
        return self.data


class Dataset(BaseDataLoader):
    """An abstract class representing a Dataset.

    All other datasets should subclass it. All subclasses should override
    ``__len__``, that provides the size of the dataset, and ``__getitem__``,
    supporting integer indexing in range from 0 to len(self) exclusive.
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __getitem__(self, index):
        """Return one sample

        index: {0, ..., len(self)-1}
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __len__(self):
        """Return the number of all samples
        """
        raise NotImplementedError

    def batch_iter(self, batch_size=32, shuffle=False, num_workers=0, drop_last=False, **kwargs):
        """Return a batch-iterator

        Arguments:
            dataset (Dataset): dataset from which to load the data.
            batch_size (int, optional): how many samples per batch to load
                (default: 1).
            shuffle (bool, optional): set to ``True`` to have the data reshuffled
                at every epoch (default: False).
            num_workers (int, optional): how many subprocesses to use for data
                loading. 0 means that the data will be loaded in the main process
                (default: 0)
            drop_last (bool, optional): set to ``True`` to drop the last incomplete batch,
                if the dataset size is not divisible by the batch size. If False and
                the size of dataset is not divisible by the batch size, then the last batch
                will be smaller. (default: False)

        Returns:
            iterator
        """
        dl = DataLoader(self, batch_size=batch_size,
                        collate_fn=numpy_collate,
                        shuffle=shuffle,
                        num_workers=num_workers,
                        drop_last=drop_last,
                        **kwargs)
        return iter(dl)

    def load_all(self, batch_size=32, num_workers=0, **kwargs):
        """Load the whole dataset into memory
        Arguments:
            batch_size (int, optional): how many samples per batch to load
                (default: 1).
            num_workers (int, optional): how many subprocesses to use for data
                loading. 0 means that the data will be loaded in the main process
                (default: 0)
        """
        return numpy_collate_concat([x for x in tqdm(self.batch_iter(batch_size,
                                                                     num_workers=num_workers))])


class BatchDataset(BaseDataLoader):
    """An abstract class representing a BatchDataset.
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __getitem__(self, index):
        """Return one batch
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __len__(self):
        """Number of all batches
        """
        raise NotImplementedError

    def batch_iter(self, num_workers=0, **kwargs):
        """Return a batch-iterator

        Arguments:
            dataset (Dataset): dataset from which to load the data.
            batch_size (int, optional): how many samples per batch to load
                (default: 1).
            shuffle (bool, optional): set to ``True`` to have the data reshuffled
                at every epoch (default: False).
            num_workers (int, optional): how many subprocesses to use for data
                loading. 0 means that the data will be loaded in the main process
                (default: 0)
            drop_last (bool, optional): set to ``True`` to drop the last incomplete batch,
                if the dataset size is not divisible by the batch size. If False and
                the size of dataset is not divisible by the batch size, then the last batch
                will be smaller. (default: False)

        Returns:
            iterator
        """
        dl = DataLoader(self, batch_size=1,
                        collate_fn=numpy_collate_concat,
                        shuffle=False,
                        num_workers=num_workers,
                        drop_last=False)
        return iter(dl)


class SampleIterator(BaseDataLoader):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __iter__(self):
        raise NotImplementedError

    # TODO - how to maintain compatibility with python2?
    @abc.abstractmethod
    def __next__(self):
        raise NotImplementedError

    next = __next__

    def batch_iter(self, batch_size=32, **kwargs):
        # TODO - implement this in parallel - add `num_workers` argument
        # https://github.com/fchollet/keras/blob/master/keras/utils/data_utils.py#L589
        l = []
        for x in iter(self):
            l.append(x)
            if len(l) == batch_size:
                ret = numpy_collate(l)
                # remove all elements
                del l[:]
                yield ret
        # Return the rest
        if len(l) > 0:
            yield numpy_collate(l)


class BatchIterator(BaseDataLoader):
    @abc.abstractmethod
    def __iter__(self):
        raise NotImplementedError

    # TODO - how to maintain compatibility with python2?
    @abc.abstractmethod
    def __next__(self):
        raise NotImplementedError

    next = __next__

    def batch_iter(self, **kwargs):
        # TODO - implement num_workers
        return iter(self)


class SampleGenerator(SampleIterator):
    """Transform a generator of samples into SampleIterator
    """
    generator_fn = None

    @classmethod
    def from_fn(cls, generator_fn):
        """setup the class variable
        """
        cls.generator_fn = staticmethod(generator_fn)
        return cls

    @classmethod
    def _get_generator_fn(cls):
        assert cls.generator_fn is not None
        return cls.generator_fn

    def __init__(self, *args, **kwargs):
        self.gen = self._get_generator_fn()(*args, **kwargs)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.gen)

    next = __next__  # python 2 compatibility


class BatchGenerator(BatchIterator):
    """Transform a generator of batches into BatchIterator
    """
    generator_fn = None

    @classmethod
    def from_fn(cls, generator_fn):
        cls.generator_fn = staticmethod(generator_fn)
        return cls

    @classmethod
    def _get_generator_fn(cls):
        assert cls.generator_fn is not None
        return cls.generator_fn

    def __init__(self, *args, **kwargs):
        self.gen = self._get_generator_fn()(*args, **kwargs)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.gen)

    next = __next__  # python 2 compatibility
# --------------------------------------------


def get_dataloader_factory(dataloader, source="kipoi"):

    # pull the dataloader & get the dataloader directory
    source = kipoi.config.get_source(source)
    yaml_path = source.pull_dataloader(dataloader)
    dataloader_dir = os.path.dirname(yaml_path)

    # Setup dataloader description
    dl = DataLoaderDescription.load(yaml_path)
    # --------------------------------------------
    # input the
    file_path, obj_name = tuple(dl.defined_as.split("::"))
    with cd(dataloader_dir):  # move to the dataloader directory temporarily
        CustomDataLoader = getattr(load_module(file_path), obj_name)

    # check that dl.type is correct
    if dl.type not in AVAILABLE_DATALOADERS:
        raise ValueError("dataloader type: {0} is not in supported dataloaders:{1}".
                         format(dl.type, list(AVAILABLE_DATALOADERS.keys())))
    # check that the extractor arguments match yaml arguments
    if not getargs(CustomDataLoader) == set(dl.args.keys()):
        raise ValueError("DataLoader arguments: \n{0}\n don't match ".format(set(getargs(CustomDataLoader))) +
                         "the specification in the dataloader.yaml file:\n{0}".
                         format(set(dl.args.keys())))
    # check that CustomDataLoader indeed interits from the right DataLoader
    if dl.type in DATALOADERS_AS_FUNCTIONS:
        # transform the functions into objects
        assert isinstance(CustomDataLoader, types.FunctionType)
        CustomDataLoader = AVAILABLE_DATALOADERS[dl.type].from_fn(CustomDataLoader)
    else:
        if not issubclass(CustomDataLoader, AVAILABLE_DATALOADERS[dl.type]):
            raise ValueError("DataLoader does't inherit from the specified dataloader: {0}".
                             format(AVAILABLE_DATALOADERS[dl.type].__name__))
    logger.info('successfully loaded the dataloader from {}'.
                format(os.path.normpath(os.path.join(dataloader_dir, dl.defined_as))))
    # Inherit the attributes from dl
    # TODO - make this more automatic / DRY
    # write a method to load those things?
    CustomDataLoader.type = dl.type
    CustomDataLoader.defined_as = dl.defined_as
    CustomDataLoader.args = dl.args
    CustomDataLoader.info = dl.info
    CustomDataLoader.output_schema = dl.output_schema
    CustomDataLoader.dependencies = dl.dependencies
    CustomDataLoader.postprocessing = dl.postprocessing
    # keep it hidden?
    CustomDataLoader._yaml_path = yaml_path
    CustomDataLoader.source = source
    # TODO - rename?
    CustomDataLoader.source_dir = dataloader_dir

    # Add init_example method
    CustomDataLoader.example_kwargs = example_kwargs(CustomDataLoader.args)

    def init_example(cls):
        return cls(**cls.example_kwargs)
    CustomDataLoader.init_example = classmethod(init_example)
    CustomDataLoader.print_args = classmethod(print_dl_kwargs)

    return CustomDataLoader


AVAILABLE_DATALOADERS = {"PreloadedDataset": PreloadedDataset,
                         "Dataset": Dataset,
                         "BatchDataset": BatchDataset,
                         "SampleIterator": SampleIterator,
                         "SampleGenerator": SampleGenerator,
                         "BatchIterator": BatchIterator,
                         "BatchGenerator": BatchGenerator}

DATALOADERS_AS_FUNCTIONS = ["PreloadedDataset", "SampleGenerator", "BatchGenerator"]



import abc
from abc import abstractmethod
import logging
import copy
import numpy as np

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())



class SequenceMutator(object):
    __metaclass__ = abc.ABCMeta

    @abstractmethod
    def __call__(self, input_set, preproc_conv, allele, s_dir):
        """
        Process sequence object `input_set` according to information given in the `preproc_conv` dataframe of which
        the column with name set in argument `allele` is used. `s_dir` defines the output sequence direction: 'fwd'
        or 'rc'. The DNA sequence will then be mutated accordingly.
        """
        raise NotImplementedError("This functionality has to be implemented in the specific subclasses")


def _modify_bases(seq_obj, lines, pos, base, is_rc, return_ref_warning=False):
    # Assumes a fixed order of ACGT and requires one-hot encoding
    assert lines.shape[0] == pos.shape[0]
    assert base.shape[0] == pos.shape[0]
    alphabet = np.array(['A', "C", "G", "T"])
    base_sel = np.where(alphabet[None, :] == base[:, None])
    base_sel_idx = (np.zeros(base.shape[0], dtype=np.int) - 1)
    base_sel_idx[base_sel[0]] = base_sel[1]
    pos = copy.deepcopy(pos)
    # get bases that are not in the alphabet
    known_bases = base_sel_idx != -1
    if is_rc.sum() != 0:
        # is_rc is given relative to the input data, so it as to be transferred back into the variant definition order.
        pos[is_rc] = seq_obj.shape[1] - pos[is_rc] - 1
        base_sel_idx[is_rc] = alphabet.shape[0] - base_sel_idx[is_rc] - 1
    warn_lines = []
    if return_ref_warning:
        wrong_ref = seq_obj[lines[known_bases], pos[known_bases], base_sel_idx[known_bases]] != 1
        warn_lines = np.array(lines[known_bases])[wrong_ref].tolist()
    # Reset the base which was there from the preprocessor
    # Reset the unknown bases just to 0
    seq_obj[lines, pos, :] = 0
    # Set the allele
    seq_obj[lines[known_bases], pos[known_bases], base_sel_idx[known_bases]] = 1
    return warn_lines

def rc_str(dna):
    """Reverse complement a string
    """
    complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N'}
    ks = list(complement.keys())
    for k in ks:
        complement[k.lower()] = complement[k].lower()
    return ''.join([complement[base] for base in dna[::-1]])

def _modify_single_string_base(seq_obj, pos, base, is_rc):
    """modify the single base of a string"""
    if is_rc:
        pos = len(seq_obj) - pos - 1
        base = rc_str(base)
    ret_obj = seq_obj[:pos] + base + seq_obj[(pos + 1):]
    return ret_obj

class OneHotSequenceMutator_OLD(SequenceMutator):
    def __init__(self, array_trafo=None):
        # from the model info object guess
        self.array_trafo = array_trafo

    def __call__(self, input_set, preproc_conv, allele, s_dir):
        """
        Process sequence object `input_set` according to information given in the `preproc_conv` dataframe of which
        the column with name set in argument `allele` is used. `s_dir` defines the output sequence direction: 'fwd'
        or 'rc'. The DNA sequence will then be mutated accordingly. `array_trafo` objects are used to bring non-
        (N,seq_len, 4) arrays into the that shape and convert it back.
        """
        # make sure the sequence objects have the correct length (acording to the ranges specifications)
        if self.array_trafo is not None:
            input_set = self.array_trafo.to_standard(input_set)
        # only keep lines that
        preproc_conv_mutate = preproc_conv.query("do_mutate")
        exp_seq_lens = (preproc_conv_mutate["end"] - preproc_conv_mutate["start"] + 1).unique()
        assert exp_seq_lens.shape[0] == 1
        assert input_set.shape[1] == exp_seq_lens[0]
        assert preproc_conv_mutate["strand"].isin(["+", "-", "*", "."]).all()
        # Modify bases according to allele
        get_warnings = allele == "ref"
        ref_warnings = _modify_bases(seq_obj=input_set,
                                     lines=preproc_conv_mutate["pp_line"].values,
                                     pos=preproc_conv_mutate["varpos_rel"].values.astype(np.int),
                                     base=preproc_conv_mutate[allele].str.upper().values,
                                     is_rc=preproc_conv_mutate["strand"].values == "-",
                                     return_ref_warning=get_warnings)
        for ppl in ref_warnings:
            pcl = np.where(preproc_conv["pp_line"].values == ppl)[0][0]
            vstr = "".join([['A', 'C', 'G', 'T', 'N'][x.argmax() if (x.sum() != 0) and
                                                      np.all(np.in1d(x, np.arange(0, 4))) else 4] for x in input_set[pcl, ...]])
            logger.warn("Variant reference allele is not the allele present in sequence for:\n%s\n"
                        "Sequence:\n%s" % (str(preproc_conv.iloc[pcl]), vstr))
        # generate reverse complement if needed
        if s_dir == "rc":
            input_set = input_set[:, ::-1, ::-1]
        if self.array_trafo is not None:
            input_set = self.array_trafo.from_standard(input_set)
        return input_set


class OneHotSequenceMutator(SequenceMutator):
    def __init__(self, array_trafo=None):
        # from the model info object guess
        self.array_trafo = array_trafo

    def __call__(self, input_set, variant_loc, allele, s_dir):
        """
        Process sequence object `input_set` according to information given in the `preproc_conv` dataframe of which
        the column with name set in argument `allele` is used. `s_dir` defines the output sequence direction: 'fwd'
        or 'rc'. The DNA sequence will then be mutated accordingly. `array_trafo` objects are used to bring non-
        (N,seq_len, 4) arrays into the that shape and convert it back.
        """
        # make sure the sequence objects have the correct length (acording to the ranges specifications)
        if self.array_trafo is not None:
            input_set = self.array_trafo.to_standard(input_set)
        # only keep lines that
        variant_loc_mutate = variant_loc.subset_to_mutate()
        exp_seq_lens = np.unique(variant_loc_mutate.get_seq_lens())
        assert exp_seq_lens.shape[0] == 1
        assert input_set.shape[1] == exp_seq_lens[0]
        assert variant_loc_mutate.strand_vals_valid()
        # Modify bases according to allele
        get_warnings = allele == "ref"
        ref_warnings = _modify_bases(seq_obj=input_set,
                                     lines=variant_loc_mutate.get("pp_line"),
                                     pos=variant_loc_mutate.get("varpos_rel").astype(np.int),
                                     base=variant_loc_mutate.get(allele, trafo = lambda x: x.upper()),
                                     is_rc=variant_loc_mutate.get("strand") == "-",
                                     return_ref_warning=get_warnings)
        for ppl in ref_warnings:
            pcl = np.where(variant_loc.get("pp_line") == ppl)[0][0]
            vstr = "".join([['A', 'C', 'G', 'T', 'N'][x.argmax() if (x.sum() != 0) and
                                                      np.all(np.in1d(x, np.arange(0, 4))) else 4] for x in input_set[pcl, ...]])
            logger.warn("Variant reference allele is not the allele present in sequence for:\n%s\n"
                        "Sequence:\n%s" % (str(variant_loc.to_df().iloc[pcl]), vstr))
        # generate reverse complement if needed
        if s_dir == "rc":
            input_set = input_set[:, ::-1, ::-1]
        if self.array_trafo is not None:
            input_set = self.array_trafo.from_standard(input_set)
        return input_set




class DNAStringSequenceMutator_OLD(SequenceMutator):
    def __init__(self, array_trafo=None):
        self.array_trafo = array_trafo

    def __call__(self, input_set, preproc_conv, allele, s_dir):
        """
        Process sequence object `input_set` according to information given in the `preproc_conv` dataframe of which
        the column with name set in argument `allele` is used. `s_dir` defines the output sequence direction: 'fwd'
        or 'rc'. The DNA sequence will then be mutated accordingly.

        The input sequence object must be according to the definition in the dataloader.
        """
        # input_set has to be <list(<str>)> which is achieved by the `array_trafo`.
        if self.array_trafo is not None:
            input_set = self.array_trafo.to_standard(input_set)
        # Every input sequence may have different length, check that ranges defition matches the input.
        seq_lens = np.array([len(el) for el in input_set])
        exp_seq_lens = np.array(preproc_conv["end"] - preproc_conv["start"] + 1)
        assert np.all(seq_lens == exp_seq_lens)
        strands = preproc_conv.query("do_mutate")["strand"]
        assert strands.isin(["+", "-", "*", "."]).all()
        # Modify bases according to allele.
        output_set = []
        for pcl, l in enumerate(preproc_conv["pp_line"].values):
            if preproc_conv["do_mutate"].values[pcl]:
                output_set.append(_modify_single_string_base(input_set[l],
                                                             pos=int(preproc_conv["varpos_rel"].values[pcl]),
                                                             base=preproc_conv[allele].values[pcl],
                                                             is_rc=preproc_conv["strand"].values[pcl] == "-"))
                if allele == "ref":
                    is_rc = preproc_conv["strand"].values[pcl] == "-"
                    base = preproc_conv[allele].values[pcl]
                    vstr = input_set[l]
                    if is_rc:
                        vstr = rc_str(input_set[l])
                    if vstr[int(preproc_conv["varpos_rel"].values[pcl])].upper() != base.upper():
                        logger.warn("Variant reference allele is not the allele present in sequence for:\n%s\n"
                                    "Sequence:\n%s" % (str(preproc_conv.iloc[pcl]), str(input_set[l])))
            else:
                output_set.append(input_set[l])
        # subset to the lines that have been identified
        if len(output_set) != preproc_conv.shape[0]:
            raise Exception("Mismatch between requested and generated DNA sequences.")
        # generate reverse complement if needed
        if s_dir == "rc":
            output_set = [rc_str(el) for el in output_set]
        if self.array_trafo is not None:
            output_set = self.array_trafo.from_standard(output_set)
        return output_set


class DNAStringSequenceMutator(SequenceMutator):
    def __init__(self, array_trafo=None):
        self.array_trafo = array_trafo

    def __call__(self, input_set, variant_loc, allele, s_dir):
        """
        Process sequence object `input_set` according to information given in the `preproc_conv` dataframe of which
        the column with name set in argument `allele` is used. `s_dir` defines the output sequence direction: 'fwd'
        or 'rc'. The DNA sequence will then be mutated accordingly.

        The input sequence object must be according to the definition in the dataloader.
        """
        # input_set has to be <list(<str>)> which is achieved by the `array_trafo`.
        if self.array_trafo is not None:
            input_set = self.array_trafo.to_standard(input_set)
        # Every input sequence may have different length, check that ranges defition matches the input.
        variant_loc_mutate = variant_loc.subset_to_mutate()
        seq_lens = np.array([len(el) for el in input_set])
        exp_seq_lens = variant_loc.get_seq_lens()
        assert np.all(seq_lens == exp_seq_lens)
        assert variant_loc_mutate.strand_vals_valid()
        # Modify bases according to allele.
        output_set = []
        for pcl, l in enumerate(variant_loc["pp_line"]):
            if variant_loc["do_mutate"][pcl]:
                output_set.append(_modify_single_string_base(input_set[l],
                                                             pos=int(variant_loc["varpos_rel"][pcl]),
                                                             base=variant_loc[allele][pcl],
                                                             is_rc=variant_loc["strand"][pcl] == "-"))
                if allele == "ref":
                    is_rc = variant_loc["strand"][pcl] == "-"
                    base = variant_loc[allele][pcl]
                    vstr = input_set[l]
                    if is_rc:
                        vstr = rc_str(input_set[l])
                    if vstr[int(variant_loc["varpos_rel"][pcl])].upper() != base.upper():
                        logger.warn("Variant reference allele is not the allele present in sequence for:\n%s\n"
                                    "Sequence:\n%s" % (str(variant_loc.to_df().iloc[pcl]), str(input_set[l])))
            else:
                output_set.append(input_set[l])
        # subset to the lines that have been identified
        if len(output_set) != variant_loc.num_entries():
            raise Exception("Mismatch between requested and generated DNA sequences.")
        # generate reverse complement if needed
        if s_dir == "rc":
            output_set = [rc_str(el) for el in output_set]
        if self.array_trafo is not None:
            output_set = self.array_trafo.from_standard(output_set)
        return output_set


"""Postprocessing CLI
"""
from __future__ import absolute_import
from __future__ import print_function

import argparse
import copy
import logging
import os

import kipoi
from kipoi.cli.parser_utils import add_model, add_dataloader, file_exists, dir_exists
from kipoi.components import default_kwargs
from kipoi.postprocessing.variant_effects.components import VarEffectFuncType
from kipoi.utils import load_module, getargs
from kipoi.utils import parse_json_file_str

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

scoring_options = {
    # deepsea_effect diff logit_diff logit_ref logit_alt
    # TODO - we should add more options to it: ref, alt, ref_rc, alt_rc
    "ref": kipoi.variant_effects.Ref,
    "alt": kipoi.variant_effects.Alt,
    "diff": kipoi.variant_effects.Diff,
    "logit_ref": kipoi.variant_effects.LogitRef,
    "logit_alt": kipoi.variant_effects.LogitAlt,
    "logit": kipoi.variant_effects.Logit,
    "deepsea_effect": kipoi.variant_effects.DeepSEA_effect
}

scoring_names = {
    VarEffectFuncType.diff: "diff",
    VarEffectFuncType.ref: "ref",
    VarEffectFuncType.alt: "alt",
    VarEffectFuncType.logit: "logit",
    VarEffectFuncType.logit_ref: "logit_ref",
    VarEffectFuncType.logit_alt: "logit_alt",
    VarEffectFuncType.deepsea_effect: "deepsea_effect",
}

# if e.g. 'logit' is allowed then also 'logit_alt' is allowed, etc. values here refer to 'scoring_options' keys.
categorical_enable = {
    "__any__": ["diff", "ref", "alt"],
    "logit": ["logit", "logit_ref", "logit_alt", "deepsea_effect"]
}

builtin_default_kwargs = {"rc_merging": "mean"}


def get_avail_scoring_methods(model):
    if model.postprocessing.variant_effects is None:
        raise Exception("Model deosn't support variant effect prediction according model yaml file.")
    avail_scoring_fns = []  # contains callables
    avail_scoring_fn_def_args = []  # default kwargs for every callable
    avail_scoring_fn_names = []  # contains the labels
    default_scoring_fns = []  # contains the labels of the defaults
    for sf in model.postprocessing.variant_effects.scoring_functions:
        if sf.type is not VarEffectFuncType.custom:
            sn = scoring_names[sf.type]
            sf_obj = scoring_options[sn]
            s_label = sn
            if (sf.name != "") and (sf.name not in scoring_options):
                if sf.name in avail_scoring_fn_names:
                    raise Exception("Scoring function names have to unique in the model yaml file.")
                s_label = sf.name
            def_kwargs = builtin_default_kwargs
        else:
            prefix = ""
            if (sf.name == "") or (sf.name in scoring_options):
                prefix = "custom_"
            s_label = prefix + sf.name
            if s_label in avail_scoring_fn_names:
                raise Exception("Scoring function names have to unique in the model yaml file.")
            if sf.defined_as == "":
                raise Exception("`defined_as` has to be defined for a custom function.")
            file_path, obj_name = tuple(sf.defined_as.split("::"))
            sf_obj = getattr(load_module(file_path), obj_name)
            # check that the scoring function arguments match yaml arguments
            if not getargs(sf_obj) == set(sf.args.keys()):
                raise ValueError("Scoring function arguments: \n{0}\n don't match ".format(set(getargs(sf_obj))) +
                                 "the specification in the dataloader.yaml file:\n{0}".
                                 format(set(sf.args.keys())))

            def_kwargs = None
            if all([(sf.args[k].default is not None) or (sf.args[k].optional) for k in sf.args]):
                def_kwargs = default_kwargs(sf.args)
            if len(sf.args) == 0:
                def_kwargs = {}  # indicates that this callable doesn't accept any arguments for initialisation.

        if s_label in avail_scoring_fn_names:
            raise Exception("Mulitple scoring functions defined with name '%s' in the model yaml file!" % s_label)

        avail_scoring_fn_def_args.append(def_kwargs)
        avail_scoring_fns.append(sf_obj)
        avail_scoring_fn_names.append(s_label)
        if sf.default:
            default_scoring_fns.append(s_label)

    # if no default scoring functions have been set then take all of the above.
    if len(default_scoring_fns) == 0:
        default_scoring_fns = copy.copy(avail_scoring_fn_names)

    # try to complete the set of functions if needed. None of the additional ones will become part of the defaults
    additional_scoring_fn_def_args =[]
    additional_scoring_fns = []
    additional_scoring_fn_names = []
    for scoring_fn, def_args in zip(avail_scoring_fns, avail_scoring_fn_def_args):
        scoring_name = None
        for sn in scoring_options:
            if scoring_fn is scoring_options[sn]:
                scoring_name = sn
                break
        if scoring_name is None:
            continue
        categories = [cat for cat in categorical_enable if scoring_name in categorical_enable[cat]]
        for cat in categories:
            for scoring_name in categorical_enable[cat]:
                if (scoring_options[scoring_name] not in avail_scoring_fns) and \
                        (scoring_options[scoring_name] not in additional_scoring_fns):
                    additional_scoring_fn_def_args.append(def_args)
                    additional_scoring_fns.append(scoring_options[scoring_name])
                    s_label = scoring_name
                    if s_label in avail_scoring_fn_names:
                        s_label = "default_" + s_label
                    additional_scoring_fn_names.append(s_label)

    avail_scoring_fns += additional_scoring_fns
    avail_scoring_fn_def_args += additional_scoring_fn_def_args
    avail_scoring_fn_names += additional_scoring_fn_names


    # add the default scoring functions if not already in there
    for scoring_name in categorical_enable["__any__"]:
        if scoring_options[scoring_name] not in avail_scoring_fns:
            avail_scoring_fn_def_args.append(builtin_default_kwargs)
            avail_scoring_fns.append(scoring_options[scoring_name])
            s_label = scoring_name
            if s_label in avail_scoring_fn_names:
                s_label = "default_" + s_label
            avail_scoring_fn_names.append(s_label)
            if len(default_scoring_fns) == 0:
                default_scoring_fns.append(s_label)

    return avail_scoring_fns, avail_scoring_fn_def_args, avail_scoring_fn_names, default_scoring_fns

def _get_scoring_fns(model, sel_scoring_labels, sel_scoring_kwargs):
    # get the scoring methods according to the model definition
    avail_scoring_fns, avail_scoring_fn_def_args, avail_scoring_fn_names, \
        default_scoring_fns = get_avail_scoring_methods(model)

    errmsg_scoring_kwargs = "When defining `--scoring_kwargs` a JSON representation of arguments (or the path of a" \
                            " file containing them) must be given for every `--scoring` function."

    dts = {}
    if len(sel_scoring_labels) >= 1:
        # Check if all scoring functions should be used:
        if sel_scoring_labels == ["all"]:
            if len(sel_scoring_kwargs) >= 1:
                raise ValueError("`--scoring_kwargs` cannot be defined in combination will `--scoring all`!")
            for arg_iter, k in enumerate(avail_scoring_fn_names):
                si = avail_scoring_fn_names.index(k)
                # get the default kwargs
                kwargs = avail_scoring_fn_def_args[si]
                if kwargs is None:
                    raise ValueError("No default kwargs for scoring function: %s"
                                     " `--scoring all` cannot be used. "
                                     "Please also define `--scoring_kwargs`." % (k))
                # instantiate the scoring fn
                dts[k] = avail_scoring_fns[si](**kwargs)
        else:
            # if -k set check that length matches with -s
            if len(sel_scoring_kwargs) >= 1:
                if not len(sel_scoring_labels) == len(sel_scoring_kwargs):
                    raise ValueError(errmsg_scoring_kwargs)
            for arg_iter, k in enumerate(sel_scoring_labels):
                # if -s set check is available for model
                if k in avail_scoring_fn_names:
                    si = avail_scoring_fn_names.index(k)
                    # get the default kwargs
                    kwargs = avail_scoring_fn_def_args[si]
                    # if the user has set scoring function kwargs then load them here.
                    if len(sel_scoring_kwargs) >= 1:
                        # all the {}s in -k replace by their defaults, if the default is None
                        # raise exception with the corrsponding scoring function label etc.
                        defined_kwargs = parse_json_file_str(sel_scoring_kwargs[si])
                        if len(defined_kwargs) != 0:
                            kwargs = defined_kwargs
                    if kwargs is None:
                        raise ValueError("No kwargs were given for scoring function %s"
                                         " with no defaults but required argmuents. "
                                         "Please also define `--scoring_kwargs`." % (k))
                    # instantiate the scoring fn
                    dts[k] = avail_scoring_fns[si](**kwargs)
                else:
                    logger.warn("Cannot choose scoring function %s. "
                                     "Model only supports: %s." % (k, str(avail_scoring_fn_names)))
    # if -s not set use all defaults
    elif len(default_scoring_fns) != 0:
        for arg_iter, k in enumerate(default_scoring_fns):
            si = avail_scoring_fn_names.index(k)
            kwargs = avail_scoring_fn_def_args[si]
            dts[k] = avail_scoring_fns[si](**kwargs)

    if len(dts) == 0:
        raise Exception("No scoring method was chosen!")

    return dts


# TODO - --output is not always required
def cli_score_variants(command, raw_args):
    """CLI interface to predict
    """
    AVAILABLE_FORMATS = ["tsv", "hdf5", "h5"]
    import pybedtools
    assert command == "score_variants"
    parser = argparse.ArgumentParser('kipoi postproc {}'.format(command),
                                     description='Predict effect of SNVs using ISM.')
    add_model(parser)
    add_dataloader(parser, with_args=True)
    parser.add_argument('-v', '--vcf_path',
                        help='Input VCF.')
    # TODO - rename path to fpath
    parser.add_argument('-a', '--out_vcf_fpath',
                        help='Output annotated VCF file path.', default=None)
    parser.add_argument('--batch_size', type=int, default=32,
                        help='Batch size to use in prediction')
    parser.add_argument("-n", "--num_workers", type=int, default=0,
                        help="Number of parallel workers for loading the dataset")
    parser.add_argument("-i", "--install_req", action='store_true',
                        help="Install required packages from requirements.txt")
    parser.add_argument('-r', '--restriction_bed', default=None,
                        help="Regions for prediction can only be subsets of this bed file")
    parser.add_argument('-o', '--output', required=False,
                        help="Additional output file. File format is inferred from the file path ending" +
                        ". Available file formats are: {0}".format(",".join(AVAILABLE_FORMATS)))
    parser.add_argument('-s', "--scoring", default="diff", nargs="+",
                        help="Scoring method to be used. Only scoring methods selected in the model yaml file are"
                             "available except for `diff` which is always available. Select scoring function by the"
                             "`name` tag defined in the model yaml file.")
    parser.add_argument('-k', "--scoring_kwargs", default="", nargs="+",
                        help="JSON definition of the kwargs for the scoring functions selected in --scoring. The "
                             "definiton can either be in JSON in the command line or the path of a .json file. The "
                             "individual JSONs are expected to be supplied in the same order as the labels defined in "
                             "--scoring. If the defaults or no arguments should be used define '{}' for that respective "
                             "scoring method.")

    args = parser.parse_args(raw_args)

    # extract args for kipoi.variant_effects.predict_snvs
    vcf_path = args.vcf_path
    out_vcf_fpath = args.out_vcf_fpath
    dataloader_arguments = parse_json_file_str(args.dataloader_args)

    # infer the file format
    args.file_format = args.output.split(".")[-1]
    if args.file_format not in AVAILABLE_FORMATS:
        logger.error("File ending: {0} for file {1} not from {2}".
                     format(args.file_format, args.output, AVAILABLE_FORMATS))
        sys.exit(1)

    if args.file_format in ["hdf5", "h5"]:
        # only if hdf5 output is used
        import deepdish

    # Check that all the folders exist
    file_exists(args.vcf_path, logger)
    dir_exists(os.path.dirname(args.out_vcf_fpath), logger)
    if args.output is not None:
        dir_exists(os.path.dirname(args.output), logger)
    # --------------------------------------------
    # install args
    if args.install_req:
        kipoi.pipeline.install_model_requirements(args.model, args.source, and_dataloaders=True)
    # load model & dataloader
    model = kipoi.get_model(args.model, args.source)

    if args.dataloader is not None:
        Dl = kipoi.get_dataloader_factory(args.dataloader, args.dataloader_source)
    else:
        Dl = model.default_dataloader

    if not os.path.exists(vcf_path):
        raise Exception("VCF file does not exist: %s" % vcf_path)

    if not isinstance(args.scoring, list):
        args.scoring = [args.scoring]

    dts = _get_scoring_fns(model, args.scoring, args.scoring_kwargs)

    # Load effect prediction related model info
    model_info = kipoi.postprocessing.variant_effects.ModelInfoExtractor(model, Dl)

    # Select the appropriate region generator
    if args.restriction_bed is not None:
        # Select the restricted SNV-centered region generator
        pbd = pybedtools.BedTool(args.restriction_bed)
        vcf_to_region = kipoi.postprocessing.variant_effects.SnvPosRestrictedRg(model_info, pbd)
        logger.info('Restriction bed file defined. Only variants in defined regions will be tested.'
                    'Only defined regions will be tested.')
    elif model_info.requires_region_definition:
        # Select the SNV-centered region generator
        vcf_to_region = kipoi.postprocessing.variant_effects.SnvCenteredRg(model_info)
        logger.info('Using variant-centered sequence generation.')
    else:
        # No regions can be defined for the given model, VCF overlap will be inferred, hence tabixed VCF is necessary
        vcf_to_region = None
        # Make sure that the vcf is tabixed
        vcf_path = kipoi.postprocessing.variant_effects.ensure_tabixed_vcf(vcf_path)
        logger.info('Dataloader does not accept definition of a regions bed-file. Only VCF-variants that lie within'
                    'produced regions can be predicted')

    if model_info.use_seq_only_rc:
        logger.info('Model SUPPORTS simple reverse complementation of input DNA sequences.')
    else:
        logger.info('Model DOES NOT support simple reverse complementation of input DNA sequences.')

    # Get a vcf output writer if needed
    if out_vcf_fpath is not None:
        logger.info('Annotated VCF will be written to %s.' % str(out_vcf_fpath))
        vcf_writer = kipoi.postprocessing.variant_effects.VcfWriter(model, vcf_path, out_vcf_fpath)
    else:
        vcf_writer = None

    keep_predictions = args.output is not None

    res = kipoi.postprocessing.variant_effects.predict_snvs(
        model,
        Dl,
        vcf_path,
        batch_size=args.batch_size,
        num_workers=args.num_workers,
        dataloader_args=dataloader_arguments,
        vcf_to_region=vcf_to_region,
        evaluation_function_kwargs={"diff_types": dts},
        sync_pred_writer=vcf_writer,
        return_predictions=keep_predictions
    )

    # tabular files
    if args.output is not None:
        if args.file_format in ["tsv"]:
            for i, k in enumerate(res):
                # Remove an old file if it is still there...
                if i == 0:
                    try:
                        os.unlink(args.output)
                    except Exception:
                        pass
                with open(args.output, "w") as ofh:
                    ofh.write("KPVEP_%s\n" % k.upper())
                    res[k].to_csv(args.output, sep="\t", mode="a")

        if args.file_format in ["hdf5", "h5"]:
            deepdish.io.save(args.output, res)

    logger.info('Successfully predicted samples')


# --------------------------------------------
# CLI commands

command_functions = {
    'score_variants': cli_score_variants,
}
commands_str = ', '.join(command_functions.keys())

parser = argparse.ArgumentParser(
    description='Kipoi model-zoo command line tool',
    usage='''kipoi postproc <command> [-h] ...

    # Available sub-commands:
    score_variants   Score variants with a kipoi model
    ''')
parser.add_argument('command', help='Subcommand to run; possible commands: {}'.format(commands_str))


def cli_main(command, raw_args):
    args = parser.parse_args(raw_args[0:1])
    if args.command not in command_functions:
        parser.print_help()
        parser.exit(
            status=1,
            message='\nCommand `{}` not found. Possible commands: {}\n'.format(
                args.command, commands_str))
    command_fn = command_functions[args.command]
    command_fn(args.command, raw_args[1:])



"""Command-line interface for kipoi ls
"""
from __future__ import absolute_import
from __future__ import print_function

import kipoi


def cli_ls(cmd, raw_args):
    """List all kipoi-induced conda environments
    """
    dtm = kipoi.list_models()
    for m in list(dtm.source.str.cat(dtm.model, sep="::")):
        print(m)



import copy
import itertools
import logging
import os
import tempfile

import numpy as np
import pandas as pd
import six
from tqdm import tqdm

from kipoi.postprocessing.variant_effects.utils.scoring_fns import Logit
from kipoi.postprocessing.variant_effects.utils import select_from_dl_batch, OutputReshaper, default_vcf_id_gen, \
    ModelInfoExtractor, BedWriter, VariantLocalisation

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

def analyse_model_preds(model, ref, alt, diff_types,
                        output_reshaper, output_filter=None, ref_rc=None, alt_rc=None, **kwargs):
    seqs = {"ref": ref, "alt": alt}
    if ref_rc is not None:
        seqs["ref_rc"] = ref_rc
    if alt_rc is not None:
        seqs["alt_rc"] = alt_rc
    if not isinstance(diff_types, dict):
        raise Exception("diff_types has to be a dictionary of callables. Keys will be used to annotate output.")
    # This is deprecated as no simple deduction of sequence length is possible anymore
    #assert np.all([np.array(_get_seq_len(ref)) == np.array(_get_seq_len(seqs[k])) for k in seqs.keys() if k != "ref"])

    # Make predictions
    preds = {}
    out_annotation = None
    for k in seqs:
        # Flatten the model output
        preds_out, pred_labels = output_reshaper.flatten(model.predict_on_batch(seqs[k]))
        if out_annotation is None:
            out_annotation = pred_labels
            output_filter = np.zeros(pred_labels.shape[0]) == 0
            # determine which outputs should be selected
            if output_filter is None:
                if output_filter.dtype == bool:
                    assert(output_filter.shape == out_annotation.shape)
                else:
                    assert np.all(np.in1d(output_filter, out_annotation))
                    output_filter = np.in1d(out_annotation, output_filter)
        # Filter outputs if required
        preds[k] = np.array(preds_out[..., output_filter])

    # Run the analysis callables
    outputs = {}
    for k in diff_types:
        outputs[k] = pd.DataFrame(diff_types[k](**preds), columns=out_annotation[output_filter])

    return outputs



def _overlap_vcf_region(vcf_obj, regions, exclude_indels=True):
    """
    Overlap a vcf with regions generated by the dataloader
    The region definition is assumed to be 0-based hence it is converted to 1-based for tabix overlaps!
    Returns VCF records
    """
    assert isinstance(regions["chr"], list) or isinstance(regions["chr"], np.ndarray)
    contained_regions = []
    vcf_records = []
    for i in range(len(regions["chr"])):
        chrom, start, end = regions["chr"][i], regions["start"][i] + 1, regions["end"][i]
        region_str = "{0}:{1}-{2}".format(chrom, start, end)
        variants = vcf_obj(region_str)
        for record in variants:
            if record.is_indel and exclude_indels:
                continue
            vcf_records.append(record)
            contained_regions.append(i)
    #
    return vcf_records, contained_regions


# For every post-processing-activated DNA output assign a sequence mutator + the name of the metadata ranges name
# In _generate_seq_sets collect all the regions that might overlap. Make an object with mutatability:
# [{vcf_record, region, sample_within_batch_id, affected_seqeunce_output_name},...]

def merge_intervals_strandaware(ranges_dict):
    """
    Perform in-silico mutagenesis on what the dataloader has returned.  

    This function has to convert the DNA regions in the model input according to ref, alt, fwd, rc and
    return a dictionary of which the keys are compliant with evaluation_function arguments

    DataLoaders that implement fwd and rc sequence output *__at once__* are not treated in any special way.
    Perform in-silico mutagenesis on what the dataloader has returned.  

    This function has to convert the DNA regions in the model input according to ref, alt, fwd, rc and
    return a dictionary of which the keys are compliant with evaluation_function arguments

    DataLoaders that implement fwd and rc sequence output *__at once__* are not treated in any special way.
    `ranges_dict`: dictionary of GenomicsRanges
    Returns unified ranges
    """
    from intervaltree import IntervalTree
    chrom_trees = {}
    for k in ranges_dict:
        assert len(ranges_dict[k]["chr"]) == 1
        chr_str = [ranges_dict[k]["chr"][0], ranges_dict[k]["strand"][0]]
        start = ranges_dict[k]["start"][0]
        end = ranges_dict[k]["end"][0]
        if chr_str not in chrom_trees:
            chrom_trees[chr_str] = IntervalTree()
        # append new region to the interval tree
        chrom_trees[chr_str][start:end] = [k]
    # merge overlapping regions and append the metadata fields
    [chrom_trees[chr_str].merge_overlaps(lambda x, y: x + y) for chr_str in chrom_trees]
    out_regions = {k: [] for k in ["chr", "start", "end", "strand"]}
    ranges_ks = []
    for chr_str in chrom_trees:
        for interval in chrom_trees[chr_str]:
            out_regions["chr"].append(chr_str[0])
            out_regions["strand"].append(chr_str[1])
            out_regions["start"].append(interval.begin)
            out_regions["end"].append(interval.end)
            ranges_ks.append(interval.data)
    return out_regions, ranges_ks

def merge_intervals(ranges_dict):
    """
    `ranges_dict`: dictionary of GenomicsRanges
    Returns unified ranges
    """
    from intervaltree import IntervalTree
    chrom_trees = {}
    for k in ranges_dict:
        assert len(ranges_dict[k]["chr"]) == 1
        chr = ranges_dict[k]["chr"][0]
        start = ranges_dict[k]["start"][0]
        end = ranges_dict[k]["end"][0]
        if chr not in chrom_trees:
            chrom_trees[chr] = IntervalTree()
        # append new region to the interval tree
        chrom_trees[chr][start:end] = [k]
    # merge overlapping regions and append the metadata fields
    [chrom_trees[chr].merge_overlaps(lambda x, y: x + y) for chr in chrom_trees]
    # convert back to GenomicRanges-compliant dictionaries
    out_regions = {k: [] for k in ["chr", "start", "end"]}
    ranges_ks = []
    for chr in chrom_trees:
        for interval in chrom_trees[chr]:
            out_regions["chr"].append(chr)
            out_regions["start"].append(interval.begin)
            out_regions["end"].append(interval.end)
            ranges_ks.append(interval.data)
    out_regions["strand"] = ["*"] * len(out_regions["chr"])
    return out_regions, ranges_ks


def get_genomicranges_line(gr_obj, i):
    return {k: v[i:(i + 1)] for k, v in gr_obj.items()}


def get_variants_in_regions_search_vcf(dl_batch, seq_to_meta, vcf_fh):
    """
    Function that overlaps metadata ranges with a vcf by merging the regions. When regions are party overlapping then
    a variant will be tagged with all sequence-fields that participated in the merged region, hence not all input
    regions might be affected by the variant.
    """
    vcf_records = []  # list of vcf records to use
    process_lines = []  # sample id within batch
    process_seq_fields = []  # sequence fields that should be mutated
    #
    meta_to_seq = {v: [k for k in seq_to_meta if seq_to_meta[k] == v] for v in seq_to_meta.values()}
    all_meta_fields = list(set(seq_to_meta.values()))
    #
    num_samples_in_batch = len(dl_batch['metadata'][all_meta_fields[0]]["chr"])
    #
    # If we should search for the overlapping VCF lines - for every sample collect all region objects
    # under the assumption that all generated sequences have the same number of samples in a batch:
    for line_id in range(num_samples_in_batch):
        # check is there is more than one metadata_field that is used:
        if len(all_meta_fields) > 1:
            # one region per meta_field
            regions_by_meta = {k: get_genomicranges_line(dl_batch['metadata'][k], line_id)
                               for k in all_meta_fields}
            # regions_unif: union across all regions. meta_field_unif_r: meta_fields, has the length of regions_unif
            regions_unif, meta_field_unif_r = merge_intervals(regions_by_meta)
        else:
            # Only one meta_field and only one line hence:
            meta_field_unif_r = [all_meta_fields]
            # Only one region:
            regions_unif = get_genomicranges_line(dl_batch['metadata'][all_meta_fields[0]], line_id)
        #
        vcf_records_here, process_lines_rel = _overlap_vcf_region(vcf_fh, regions_unif)
        #
        for rec, sub_line_id in zip(vcf_records_here, process_lines_rel):
            vcf_records.append(rec)
            process_lines.append(line_id)
            metas = []
            for f in meta_field_unif_r[sub_line_id]:
                metas += meta_to_seq[f]
            process_seq_fields.append(metas)
    return vcf_records, process_lines, process_seq_fields


def get_variants_in_regions_sequential_vcf(dl_batch, seq_to_meta, vcf_fh, vcf_id_generator_fn):
    vcf_records = []  # list of vcf records to use
    process_ids = []  # id from genomic ranges metadata
    process_lines = []  # sample id within batch
    process_seq_fields = []  # sequence fields that should be mutated

    all_mut_seq_fields = list(set(seq_to_meta.keys()))
    all_meta_fields = list(set(seq_to_meta.values()))

    ranges = None
    for meta_field in all_meta_fields:
        rng = dl_batch['metadata'][meta_field]
        if ranges is None:
            ranges = rng
        else:
            # for k in ["chr", "start", "end", "id"]:
            for k in ["id"]:
                assert np.all(ranges[k] == rng[k])

    # Now continue going sequentially through the vcf assigning vcf record with regions to test.
    for i, returned_id in enumerate(ranges["id"]):
        for record in vcf_fh:
            id = vcf_id_generator_fn(record)
            if str(id) == str(returned_id):
                vcf_records.append(record)
                process_ids.append(returned_id)
                process_lines.append(i)
                process_seq_fields.append(all_mut_seq_fields)
                break
            else:
                # Warn here...
                logger.warn("Skipping VCF line (%s) because generated region is for different variant.." % str(id))
                pass
    return vcf_records, process_lines, process_seq_fields, process_ids



def get_variants_df(seq_key, ranges_input_obj, vcf_records, process_lines, process_ids, process_seq_fields):
    preproc_conv = {"pp_line": [], "varpos_rel": [], "ref": [], "alt": [], "start": [], "end": [], "id": [],
                    "do_mutate": []}

    if ("strand" in ranges_input_obj) and (isinstance(ranges_input_obj["strand"], list) or
                                           isinstance(ranges_input_obj["strand"], np.ndarray)):
        preproc_conv["strand"] = []

    for i, record in enumerate(vcf_records):
        assert not record.is_indel  # Catch indels, that needs a slightly modified processing
        ranges_input_i = process_lines[i]
        new_vals = {k: np.nan for k in preproc_conv.keys() if k not in ["do_mutate", "pp_line"]}
        new_vals["do_mutate"] = False
        new_vals["pp_line"] = i
        new_vals["id"] = str(process_ids[i])
        if seq_key in process_seq_fields[i]:
            pre_new_vals = {}
            pre_new_vals["start"] = ranges_input_obj["start"][ranges_input_i] + 1
            pre_new_vals["end"] = ranges_input_obj["end"][ranges_input_i]
            pre_new_vals["varpos_rel"] = int(record.POS) - pre_new_vals["start"]
            if not ((pre_new_vals["varpos_rel"] < 0) or
                    (pre_new_vals["varpos_rel"] > (pre_new_vals["end"] - pre_new_vals["start"] + 1))):

                # If variant lies in the region then continue
                pre_new_vals["do_mutate"] = True
                pre_new_vals["ref"] = str(record.REF)
                pre_new_vals["alt"] = str(record.ALT[0])

                if "strand" in preproc_conv:
                    pre_new_vals["strand"] = ranges_input_obj["strand"][ranges_input_i]

                # overwrite the nans with actual data now that
                for k in pre_new_vals:
                    new_vals[k] = pre_new_vals[k]

        for k in new_vals:
            preproc_conv[k].append(new_vals[k])

    # If strand wasn't a list then try to still fix it..
    if "strand" not in preproc_conv:
        if "strand" not in ranges_input_obj:
            preproc_conv["strand"] = ["*"] * len(preproc_conv["pp_line"])
        elif isinstance(ranges_input_obj["strand"], six.string_types):
            preproc_conv["strand"] = [ranges_input_obj["strand"]] * len(preproc_conv["pp_line"])
        else:
            raise Exception("Strand defintion invalid in metadata returned by dataloader.")

    preproc_conv_df = pd.DataFrame(preproc_conv)
    return preproc_conv_df


class SampleCounter():
    def __init__(self):
        self.sample_it_counter = 0

    def get_ids(self, number=1):
        ret = [str(i) for i in range(self.sample_it_counter, self.sample_it_counter + number)]
        self.sample_it_counter += number
        return ret


def _generate_seq_sets(dl_ouput_schema, dl_batch, vcf_fh, vcf_id_generator_fn, seq_to_mut, seq_to_meta,
                       sample_counter, vcf_search_regions=False, generate_rc=True):
    """
        Perform in-silico mutagenesis on what the dataloader has returned.  

        This function has to convert the DNA regions in the model input according to ref, alt, fwd, rc and
        return a dictionary of which the keys are compliant with evaluation_function arguments

        DataLoaders that implement fwd and rc sequence output *__at once__* are not treated in any special way.

        Arguments:
        `dataloader`: dataloader object
        `dl_batch`: model input as generated by the datalaoder
        `vcf_fh`: cyvcf2 file handle
        `vcf_id_generator_fn`: function that generates ids for VCF records
        `seq_to_mut`: dictionary that contains DNAMutator classes with seq_fields as keys
        `seq_to_meta`: dictionary that contains Metadata key names with seq_fields as keys
        `vcf_search_regions`: if `False` assume that the regions are labelled and only test variants/region combinations for
        which the label fits. If `True` find all variants overlapping with all regions and test all.
        `generate_rc`: generate also reverse complement sequences. Only makes sense if supported by model.
        """

    all_meta_fields = list(set(seq_to_meta.values()))

    num_samples_in_batch = len(dl_batch['metadata'][all_meta_fields[0]]["chr"])

    metadata_ids = sample_counter.get_ids(num_samples_in_batch)

    if "_id" in dl_batch['metadata']:
        metadata_ids = dl_batch['metadata']['id']
        assert num_samples_in_batch == len(metadata_ids)

    # now get the right region from the vcf:
    # list of vcf records to use: vcf_records
    process_ids = None  # id from genomic ranges metadata: process_lines
    # sample id within batch: process_lines
    # sequence fields that should be mutated: process_seq_fields

    if vcf_search_regions:
        vcf_records, process_lines, process_seq_fields = get_variants_in_regions_search_vcf(dl_batch, seq_to_meta, vcf_fh)
    else:
        # vcf_search_regions == False means: rely completely on the variant id
        # so for every sample assert that all metadata ranges ids agree and then find the entry.
        vcf_records, process_lines, process_seq_fields, process_ids = get_variants_in_regions_sequential_vcf(dl_batch, seq_to_meta,
                                                                                                             vcf_fh,
                                                                                                             vcf_id_generator_fn)

    # short-cut if no sequences are left
    if len(process_lines) == 0:
        return None

    if process_ids is None:
        process_ids = []
        for line_id in process_lines:
            process_ids.append(metadata_ids[line_id])

    # Generate 4 copies of the input set. subset datapoints if needed.
    input_set = {}
    seq_dirs = ["fwd"]
    if generate_rc:
        seq_dirs = ["fwd", "rc"]
    for s_dir, allele in itertools.product(seq_dirs, ["ref", "alt"]):
        k = "%s_%s" % (s_dir, allele)
        ds = dl_batch['inputs']
        all_lines = list(range(num_samples_in_batch))
        if process_lines != all_lines:
            # subset or rearrange elements
            ds = select_from_dl_batch(dl_batch['inputs'], process_lines, num_samples_in_batch)
        input_set[k] = copy.deepcopy(ds)

    # input_set matrices now are in the order required for mutation

    all_mut_seq_keys = list(set(itertools.chain.from_iterable(process_seq_fields)))

    # Start from the sequence inputs mentioned in the model.yaml
    for seq_key in all_mut_seq_keys:
        ranges_input_obj = dl_batch['metadata'][seq_to_meta[seq_key]]
        #
        # Assemble variant modification information
        #variants_df = get_variants_df(seq_key, ranges_input_obj, vcf_records,
        #                              process_lines, process_ids, process_seq_fields)

        vl = VariantLocalisation()
        vl.append_multi(seq_key, ranges_input_obj, vcf_records,
                                      process_lines, process_ids, process_seq_fields)

        # for the individual sequence input key get the correct sequence mutator callable
        dna_mutator = seq_to_mut[seq_key]

        # Actually modify sequences according to annotation
        # two for loops
        for s_dir, allele in itertools.product(seq_dirs, ["ref", "alt"]):
            k = "%s_%s" % (s_dir, allele)
            if isinstance(dl_ouput_schema.inputs, dict):
                if seq_key not in input_set[k]:
                    raise Exception("Sequence field %s is missing in DataLoader output!" % seq_key)
                input_set[k][seq_key] = dna_mutator(input_set[k][seq_key], vl, allele, s_dir)
            elif isinstance(dl_ouput_schema.inputs, list):
                modified_set = []
                for seq_el, input_schema_el in zip(input_set[k], dl_ouput_schema.inputs):
                    if input_schema_el.name == seq_key:
                        modified_set.append(dna_mutator(seq_el, vl, allele, s_dir))
                    else:
                        modified_set.append(seq_el)
                input_set[k] = modified_set
            else:
                input_set[k] = dna_mutator(input_set[k], vl, allele, s_dir)

    #
    # Reformat so that effect prediction function will get its required inputs
    pred_set = {"ref": input_set["fwd_ref"], "alt": input_set["fwd_alt"]}
    if generate_rc:
        pred_set["ref_rc"] = input_set["rc_ref"]
        pred_set["alt_rc"] = input_set["rc_alt"]
    pred_set["line_id"] = np.array(process_ids).astype(str)
    pred_set["vcf_records"] = vcf_records
    return pred_set



def predict_snvs(model,
                 dataloader,
                 vcf_fpath,
                 batch_size,
                 num_workers=0,
                 dataloader_args=None,
                 vcf_to_region=None,
                 vcf_id_generator_fn=default_vcf_id_gen,
                 evaluation_function=analyse_model_preds,
                 evaluation_function_kwargs={'diff_types': {'logit': Logit()}},
                 sync_pred_writer=None,
                 use_dataloader_example_data=False,
                 return_predictions=False,
                 generated_seq_writer=None
                 ):
    """Predict the effect of SNVs

            Prediction of effects of SNV based on a VCF. If desired the VCF can be stored with the predicted values as
            annotation. For a detailed description of the requirements in the yaml files please take a look at
            kipoi/nbs/variant_effect_prediction.ipynb.

            # Arguments
                model: A kipoi model handle generated by e.g.: kipoi.get_model()
                dataloader: Dataloader factory generated by e.g.: kipoi.get_dataloader_factory()
                vcf_fpath: Path of the VCF defining the positions that shall be assessed. Only SNVs will be tested.
                batch_size: Prediction batch size used for calling the data loader. Each batch will be generated in 4
                    mutated states yielding a system RAM consumption of >= 4x batch size.
                num_workers: Number of parallel workers for loading the dataset.
                dataloader_args: arguments passed on to the dataloader for sequence generation, arguments
                    mentioned in dataloader.yaml > postprocessing > variant_effects > bed_input will be overwritten
                    by the methods here.
                vcf_to_region: Callable that generates a region compatible with dataloader/model from a cyvcf2 record
                vcf_id_generator_fn: Callable that generates a unique ID from a cyvcf2 record
                evaluation_function: effect evaluation function. Default is `analyse_model_preds`, which will get
                    arguments defined in `evaluation_function_kwargs`
                evaluation_function_kwargs: kwargs passed on to `evaluation_function`.
                sync_pred_writer: Single writer or list of writer objects like instances of `VcfWriter`. This object
                    will be called after effect prediction of a batch is done.
                use_dataloader_example_data: Fill out the missing dataloader arguments with the example values given in the
                    dataloader.yaml.
                return_predictions: Return all variant effect predictions as a dictionary. Setting this to False will
                    help maintain a low memory profile and is faster as it avoids concatenating batches after prediction.
                generated_seq_writer: Single writer or list of writer objects like instances of `SyncHdf5SeqWriter`.
                    This object will be called after the DNA sequence sets have been generated. If this parameter is
                    not None, no prediction will be performed and only DNA sequence will be written!! This is relevant
                    if you want to use the `predict_snvs` to generate appropriate input DNA sequences for your model.

            # Returns
                If return_predictions: Dictionary which contains a pandas DataFrame containing the calculated values
                    for each model output (target) column VCF SNV line. If return_predictions == False, returns None.
            """
    import cyvcf2
    model_info_extractor = ModelInfoExtractor(model_obj=model, dataloader_obj=dataloader)

    # If then where do I have to put my bed file in the command?

    exec_files_bed_keys = model_info_extractor.get_exec_files_bed_keys()
    temp_bed3_file = None

    vcf_search_regions = True

    # If there is a field for putting the a postprocessing bed file, then generate the bed file.
    if exec_files_bed_keys is not None:
        if vcf_to_region is not None:
            vcf_search_regions = False

            temp_bed3_file = tempfile.mktemp()  # file path of the temp file

            vcf_fh = cyvcf2.VCF(vcf_fpath, "r")

            with BedWriter(temp_bed3_file) as ofh:
                for record in vcf_fh:
                    if not record.is_indel:
                        region = vcf_to_region(record)
                        id = vcf_id_generator_fn(record)
                        for chrom, start, end in zip(region["chrom"], region["start"], region["end"]):
                            ofh.append_interval(chrom=chrom, start=start, end=end, id=id)

            vcf_fh.close()
    else:
        if vcf_to_region is not None:
            logger.warn("`vcf_to_region` will be ignored as it was set, but the dataloader does not define "
                        "a bed_input in dataloader.yaml: "
                        "postprocessing > variant_effects > bed_input.")
    # Assemble the paths for executing the dataloader
    if dataloader_args is None:
        dataloader_args = {}

    # Copy the missing arguments from the example arguments.
    if use_dataloader_example_data:
        for k in dataloader.example_kwargs:
            if k not in dataloader_args:
                dataloader_args[k] = dataloader.example_kwargs[k]

    # If there was a field for dumping the region definition bed file, then use it.
    if (exec_files_bed_keys is not None) and (not vcf_search_regions):
        for k in exec_files_bed_keys:
            dataloader_args[k] = temp_bed3_file

    model_out_annotation = model_info_extractor.get_model_out_annotation()

    out_reshaper = OutputReshaper(model.schema.targets)

    res = []

    it = dataloader(**dataloader_args).batch_iter(batch_size=batch_size,
                                                  num_workers=num_workers)

    # organise the writers in a list
    if sync_pred_writer is not None:
        if not isinstance(sync_pred_writer, list):
            sync_pred_writer = [sync_pred_writer]

    # organise the prediction writers
    if generated_seq_writer is not None:
        if not isinstance(generated_seq_writer, list):
            generated_seq_writer = [generated_seq_writer]

    # Open vcf again
    vcf_fh = cyvcf2.VCF(vcf_fpath, "r")

    # pre-process regions
    keys = set()  # what is that?

    sample_counter = SampleCounter()

    # open the writers if possible:
    if sync_pred_writer is not None:
        [el.open() for el in  sync_pred_writer if hasattr(el, "open")]

    # open seq writers if possible:
    if generated_seq_writer is not None:
        [el.open() for el in  generated_seq_writer if hasattr(el, "open")]

    for i, batch in enumerate(tqdm(it)):
        # For debugging
        # if i >= 10:
        #     break
        # becomes noticable for large vcf's. Is there a way to avoid it? (i.e. to exploit the iterative nature of dataloading)
        seq_to_mut = model_info_extractor.seq_input_mutator
        seq_to_meta = model_info_extractor.seq_input_metadata
        eval_kwargs = _generate_seq_sets(dataloader.output_schema, batch, vcf_fh, vcf_id_generator_fn,
                                         seq_to_mut=seq_to_mut, seq_to_meta=seq_to_meta,
                                         sample_counter=sample_counter, vcf_search_regions=vcf_search_regions,
                                         generate_rc=model_info_extractor.use_seq_only_rc)
        if eval_kwargs is None:
            # No generated datapoint overlapped any VCF region
            continue

        if generated_seq_writer is not None:
            for writer in generated_seq_writer:
                writer(eval_kwargs)
            # Assume that we don't actually want the predictions to be calculated...
            continue

        if evaluation_function_kwargs is not None:
            assert isinstance(evaluation_function_kwargs, dict)
            for k in evaluation_function_kwargs:
                eval_kwargs[k] = evaluation_function_kwargs[k]

        eval_kwargs["out_annotation_all_outputs"] = model_out_annotation

        res_here = evaluation_function(model, output_reshaper=out_reshaper, **eval_kwargs)
        for k in res_here:
            keys.add(k)
            res_here[k].index = eval_kwargs["line_id"]
        # write the predictions synchronously
        if sync_pred_writer is not None:
            for writer in sync_pred_writer:
                writer(res_here, eval_kwargs["vcf_records"], eval_kwargs["line_id"])
        if return_predictions:
            res.append(res_here)

    vcf_fh.close()

    # open the writers if possible:
    if sync_pred_writer is not None:
        [el.close() for el in sync_pred_writer if hasattr(el, "close")]

    # open seq writers if possible:
    if generated_seq_writer is not None:
        [el.close() for el in generated_seq_writer if hasattr(el, "close")]

    try:
        if temp_bed3_file is not None:
            os.unlink(temp_bed3_file)
    except:
        pass

    if return_predictions:
        res_concatenated = {}
        for k in keys:
            res_concatenated[k] = pd.concat([batch[k]
                                             for batch in res
                                             if k in batch])
        return res_concatenated

    return None




from .dna_reshapers import OutputReshaper, ReshapeDna, ReshapeDnaString
from .generic import select_from_dl_batch, default_vcf_id_gen, ModelInfoExtractor, SnvCenteredRg, \
    SnvPosRestrictedRg, ensure_tabixed_vcf, VariantLocalisation
from .io import VcfWriter, BedWriter, SyncHdf5SeqWriter
